@@ -59,6 +59,14 @@
             [puppetlabs.puppetdb.jdbc :as jdbc :refer [query-to-vec]]
             [puppetlabs.puppetdb.config :as conf]))
 
+;; taken from storage.clj; preserved here in case of change
+(defn insert-records*
+  "Nil/empty safe insert-records, see java.jdbc's insert-records for more "
+  [table
+   record-coll]
+  (when (seq record-coll)
+    (apply jdbc/insert! table record-coll)))
+
 (defn init-through-2-3-8
   []
 
@@ -439,35 +447,6 @@
     "ALTER TABLE ONLY resource_params
         ADD CONSTRAINT resource_params_resource_fkey FOREIGN KEY (resource) REFERENCES resource_params_cache(resource) ON DELETE CASCADE"))
 
-(def add-factset-id-fk-constraint-cmd
-  "alter table facts add constraint factset_id_fk
-     foreign key (factset_id) references factsets(id)
-     on update cascade on delete cascade")
-
-(def add-resource-events-report-id-fkey-constraint-cmd
-  "alter table resource_events add constraint resource_events_report_id_fkey
-     foreign key (report_id) references reports(id) on delete cascade")
-
-(def add-catalogs-certname-fkey-constraint-cmd
-  "alter table catalogs add constraint catalogs_certname_fkey
-     foreign key (certname) references certnames(certname)
-     on update no action on delete cascade")
-
-(def add-factsets-certname-fk-constraint-cmd
-  "alter table factsets add constraint factsets_certname_fk
-     foreign key (certname) references certnames(certname)
-     on update cascade on delete cascade")
-
-(def add-reports-certname-fk-constraint-cmd
-  "alter table reports add constraint reports_certname_fkey
-     foreign key (certname) references certnames(certname)
-     on delete cascade")
-
-(def add-certnames-reports-id-fkey-constraint-cmd
-  "alter table certnames add constraint certnames_reports_id_fkey
-     foreign key (latest_report_id) references reports(id)
-     on delete set null")
-
 (defn version-2yz-to-300-migration
   ;; This migration includes:
   ;;   Insertion of the factsets hash column
@@ -765,17 +744,17 @@
       "ALTER TABLE factsets ADD CONSTRAINT factsets_environment_id_fk
        FOREIGN KEY (environment_id) REFERENCES environments(id)
        ON UPDATE RESTRICT ON DELETE RESTRICT"
-
-      add-factset-id-fk-constraint-cmd
-
+      "ALTER TABLE facts ADD CONSTRAINT factset_id_fk
+       FOREIGN KEY (factset_id) REFERENCES factsets(id)
+       ON UPDATE CASCADE ON DELETE CASCADE"
       "ALTER TABLE factsets ADD CONSTRAINT factsets_certname_idx
          UNIQUE (certname)"
       "ALTER TABLE factsets ADD CONSTRAINT factsets_hash_key UNIQUE (hash)"
 
       "ALTER TABLE resource_events ADD CONSTRAINT resource_events_unique
          UNIQUE (report_id, resource_type, resource_title, property)"
-
-      add-resource-events-report-id-fkey-constraint-cmd
+      "ALTER TABLE resource_events ADD CONSTRAINT resource_events_report_id_fkey
+         FOREIGN KEY (report_id) REFERENCES reports(id) ON DELETE CASCADE"
 
       "ALTER TABLE certnames ADD CONSTRAINT certnames_pkey
          PRIMARY KEY (certname)"
@@ -784,11 +763,19 @@
       "ALTER TABLE edges ADD CONSTRAINT edges_certname_fkey
          FOREIGN KEY (certname) REFERENCES certnames(certname)
          ON UPDATE NO ACTION ON DELETE CASCADE"
+      "ALTER TABLE catalogs ADD CONSTRAINT catalogs_certname_fkey
+         FOREIGN KEY (certname) REFERENCES certnames(certname)
+         ON UPDATE NO ACTION ON DELETE CASCADE"
+      "ALTER TABLE factsets ADD CONSTRAINT factsets_certname_fk
+         FOREIGN KEY (certname) REFERENCES certnames(certname)
+         ON UPDATE CASCADE ON DELETE CASCADE"
+      "ALTER TABLE reports ADD CONSTRAINT reports_certname_fkey
+         FOREIGN KEY (certname) REFERENCES certnames(certname)
+         ON DELETE CASCADE"
 
-      add-catalogs-certname-fkey-constraint-cmd
-      add-factsets-certname-fk-constraint-cmd
-      add-reports-certname-fk-constraint-cmd
-      add-certnames-reports-id-fkey-constraint-cmd)))
+      "ALTER TABLE certnames ADD CONSTRAINT certnames_reports_id_fkey
+         FOREIGN KEY (latest_report_id) REFERENCES reports(id)
+         ON DELETE SET NULL")))
 
 (defn add-expired-to-certnames
   "Add a 'expired' column to the 'certnames' table, to track
@@ -881,14 +868,8 @@
    "ALTER TABLE catalog_resources RENAME TO catalog_resources_tmp"
    ;; CREATE certnames and catalog_resources transform tables
    "CREATE TABLE catalog_resources (LIKE catalog_resources_tmp INCLUDING ALL)"
-   ;; Make sure to update the constraints in
-   ;; purge-deactivated-and-expired-nodes-new! if you change
-   ;; the columns here.
    "CREATE TABLE latest_catalogs (catalog_id BIGINT NOT NULL UNIQUE REFERENCES catalogs(id) ON DELETE CASCADE, certname_id BIGINT PRIMARY KEY REFERENCES certnames(id) ON DELETE CASCADE)"
    "ALTER TABLE catalog_resources DROP COLUMN catalog_id"
-   ;; Make sure to update the constraint in
-   ;; purge-deactivated-and-expired-nodes-new! if you change
-   ;; the columns here.
    "ALTER TABLE catalog_resources ADD COLUMN certname_id BIGINT NOT NULL REFERENCES certnames(id) ON DELETE CASCADE"
    "ALTER TABLE catalog_resources ADD PRIMARY KEY (certname_id, type, title)"
 
@@ -978,11 +959,19 @@
    "ALTER TABLE certnames DROP CONSTRAINT certnames_pkey CASCADE"
    "DROP TABLE certnames"
    "ALTER TABLE certnames_transform RENAME to certnames"
+   "ALTER TABLE catalogs ADD CONSTRAINT catalogs_certname_fkey
+     FOREIGN KEY (certname) REFERENCES certnames(certname)
+     ON UPDATE NO ACTION ON DELETE CASCADE"
+   "ALTER TABLE factsets ADD CONSTRAINT factsets_certname_fk
+     FOREIGN KEY (certname) REFERENCES certnames(certname)
+     ON UPDATE CASCADE ON DELETE CASCADE"
+   "ALTER TABLE reports ADD CONSTRAINT reports_certname_fkey
+     FOREIGN KEY (certname)
+     REFERENCES certnames(certname) ON DELETE CASCADE"
 
-   add-catalogs-certname-fkey-constraint-cmd
-   add-factsets-certname-fk-constraint-cmd
-   add-reports-certname-fk-constraint-cmd
-   add-certnames-reports-id-fkey-constraint-cmd))
+   "ALTER TABLE certnames ADD CONSTRAINT certnames_reports_id_fkey
+      FOREIGN KEY (latest_report_id)
+      REFERENCES reports(id) ON DELETE SET NULL"))
 
 (defn add-certname-id-to-resource-events
   []
@@ -1032,8 +1021,8 @@
      ON resource_events(resource_title)"
    "CREATE INDEX resource_events_status_idx ON resource_events(status)"
    "CREATE INDEX resource_events_timestamp_idx ON resource_events(timestamp)"
-
-   add-resource-events-report-id-fkey-constraint-cmd))
+   "ALTER TABLE resource_events ADD CONSTRAINT resource_events_report_id_fkey
+     FOREIGN KEY (report_id) REFERENCES reports(id) ON DELETE CASCADE"))
 
 (defn add-catalog-uuid-to-reports-and-catalogs
   []
@@ -1048,6 +1037,138 @@
   (jdbc/do-commands
     "CREATE INDEX idx_certnames_latest_report_id on certnames(latest_report_id)"))
 
+(defn add-producer-to-reports-catalogs-and-factsets
+  []
+  (jdbc/do-commands
+    (sql/create-table-ddl
+      :producers
+      ["id" "bigint PRIMARY KEY"]
+      ["name" "text NOT NULL UNIQUE"])
+    "CREATE SEQUENCE producers_id_seq CYCLE"
+    "ALTER TABLE producers ALTER COLUMN id SET DEFAULT nextval('producers_id_seq')"
+    "ALTER TABLE reports ADD COLUMN producer_id bigint"
+    "ALTER TABLE factsets ADD COLUMN producer_id bigint"
+    "ALTER TABLE catalogs ADD COLUMN producer_id bigint"
+    "ALTER TABLE reports
+        ADD CONSTRAINT reports_prod_fkey FOREIGN KEY (producer_id) REFERENCES producers(id)"
+    "ALTER TABLE factsets
+        ADD CONSTRAINT factsets_prod_fk FOREIGN KEY (producer_id) REFERENCES producers(id)"
+    "ALTER TABLE catalogs
+        ADD CONSTRAINT catalogs_prod_fkey FOREIGN KEY (producer_id) REFERENCES producers(id)"
+    "CREATE INDEX idx_reports_prod ON reports(producer_id)"
+    "CREATE INDEX idx_factsets_prod ON factsets(producer_id)"
+    "CREATE INDEX idx_catalogs_prod ON catalogs(producer_id)"))
+
+(defn drop-certnames-latest-id-index
+  []
+  (jdbc/do-commands
+    "DROP INDEX idx_certnames_latest_report_id"))
+
+(defn add-noop-pending-to-reports
+  []
+  (jdbc/do-commands
+    "ALTER TABLE reports ADD COLUMN noop_pending boolean"
+    "CREATE INDEX idx_reports_noop_pending on reports using btree (noop_pending) where (noop_pending = true)"))
+
+(defn add-corrective-change-columns
+  []
+  (jdbc/do-commands
+    "alter table reports add column corrective_change boolean"
+    "alter table resource_events add column corrective_change boolean"))
+
+(defn remove-historical-catalogs
+  []
+  (jdbc/do-commands
+    "alter table catalogs drop column edges"
+    "alter table catalogs drop column resources"
+    "delete from catalogs where id not in (select catalog_id from latest_catalogs)"
+    "drop table latest_catalogs"))
+
+(defn migrate-through-app
+  "stream data from table1 through the application and into table2, applying
+   munge-fn to each row"
+  [table1 table2 column-list batchsize munge-fn]
+  (let [columns (string/join "," column-list)]
+    (jdbc/query-with-resultset
+      [(format "select %s from %s" columns (name table1))]
+      (fn [rs]
+        (->> (sql/result-set-seq rs)
+             (map munge-fn)
+             (partition-all batchsize)
+             (map #(insert-records* (name table2) %))
+             dorun)))))
+
+(defn resource-params-cache-parameters-to-jsonb
+  []
+  (jdbc/do-commands
+    (sql/create-table-ddl :resource_params_cache_transform
+                          ["resource" "bytea NOT NULL"]
+                          ["parameters" "jsonb"]))
+
+  (migrate-through-app
+    :resource_params_cache
+    :resource_params_cache_transform
+    ["encode(resource::bytea, 'hex') as resource" "parameters"]
+    500
+    #(-> %
+         (update :parameters (comp sutils/munge-jsonb-for-storage json/parse-string))
+         (update :resource sutils/munge-hash-for-storage)))
+
+  (jdbc/do-commands
+    "alter table catalog_resources drop constraint catalog_resources_resource_fkey"
+    "alter table resource_params drop constraint resource_params_resource_fkey"
+    "drop table resource_params_cache"
+
+    "alter table resource_params_cache_transform rename to resource_params_cache"
+
+    "alter table resource_params_cache add constraint resource_params_cache_pkey
+     primary key (resource)"
+    "alter table catalog_resources add constraint catalog_resources_resource_fkey
+     foreign key (resource) references resource_params_cache(resource) on delete cascade"
+    "alter table resource_params add constraint resource_params_resource_fkey
+     foreign key (resource) references resource_params_cache(resource) on delete cascade"
+    "create index rpc_hash_expr_idx on resource_params_cache(encode(resource, 'hex'))")   )
+
+(defn fact-values-value-to-jsonb
+  []
+  (jdbc/do-commands
+    (sql/create-table-ddl :fact_values_transform
+                          ["id" "bigint NOT NULL PRIMARY KEY DEFAULT nextval('fact_values_id_seq')"]
+                          ["value_hash" "bytea NOT NULL UNIQUE"]
+                          ["value_type_id" "bigint NOT NULL"]
+                          ["value_integer" "bigint"]
+                          ["value_float" "double precision"]
+                          ["value_string" "text"]
+                          ["value_boolean" "boolean"]
+                          ["value" "jsonb"]))
+
+  (migrate-through-app
+    :fact_values
+    :fact_values_transform
+    ["id" "encode(value_hash::bytea, 'hex') as value_hash" "value_type_id"
+     "value_integer" "value_float" "value_string" "value_boolean" "value"]
+    500
+    #(-> %
+         (update :value (comp sutils/munge-jsonb-for-storage json/parse-string))
+         (update :value_hash sutils/munge-hash-for-storage)))
+
+  (jdbc/do-commands
+    "alter table facts drop constraint fact_value_id_fk"
+
+    "drop table fact_values"
+    "alter table fact_values_transform rename to fact_values"
+
+    "alter table fact_values rename constraint fact_values_transform_value_hash_key
+     to fact_values_value_hash_key"
+
+    "alter table fact_values rename constraint fact_values_transform_pkey
+     to fact_values_pkey"
+
+    "alter table facts add constraint fact_value_id_fk foreign key
+     (fact_value_id) references fact_values(id) on update restrict on delete restrict"
+    "create index fact_values_value_float_idx on fact_values(value_float)"
+    "create index fact_values_value_integer_idx on fact_values(value_integer)"))
+
 (def migrations
   "The available migrations, as a map from migration version to migration function."
   {28 init-through-2-3-8
@@ -1070,7 +1191,15 @@
    42 add-support-for-historical-catalogs
    43 add-indexes-for-reports-summary-query
    44 add-catalog-uuid-to-reports-and-catalogs
-   45 index-certnames-latest-report-id})
+   45 index-certnames-latest-report-id
+   46 drop-certnames-latest-id-index
+   47 add-producer-to-reports-catalogs-and-factsets
+   48 add-noop-pending-to-reports
+   49 add-corrective-change-columns
+   50 remove-historical-catalogs
+   51 fact-values-value-to-jsonb
+   52 resource-params-cache-parameters-to-jsonb})
+
 
 (def desired-schema-version (apply max (keys migrations)))
 
