@@ -299,6 +299,21 @@
   (when env-name
     (ensure-row :environments {:environment env-name})))
 
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;; Producers querying/updating
+
+(pls/defn-validated producer-id :- (s/maybe s/Int)
+  "Returns the id (primary key) from the producers table for the given `prod-name`"
+  [prod-name :- s/Str]
+  (query-id :producers {:name prod-name}))
+
+(pls/defn-validated ensure-producer :- (s/maybe s/Int)
+  "Check if the given `prod-name` exists, creates it if it does not. Always returns
+   the id of the `prod-name` (whether created or existing)"
+  [prod-name :- (s/maybe s/Str)]
+  (when prod-name
+    (ensure-row :producers {:name prod-name})))
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Status querying/updating
 
@@ -323,27 +338,15 @@
   [certname]
   {:pre [certname]}
   (jdbc/query-with-resultset
-   [(format (str "SELECT catalogs.id AS catalog_id, certnames.id AS certname_id, "
-                 " %s AS catalog_hash, catalogs.producer_timestamp"
-                 " FROM certnames "
-                 " LEFT JOIN latest_catalogs ON certnames.id = latest_catalogs.certname_id"
-                 " LEFT JOIN catalogs ON catalogs.id = latest_catalogs.catalog_id"
-                 " WHERE certnames.certname=?")
+   [(format "select catalogs.id as catalog_id, certnames.id as certname_id,
+             %s as catalog_hash, catalogs.producer_timestamp
+             from certnames
+             left join catalogs on catalogs.certname=certnames.certname
+             where certnames.certname=?"
             (sutils/sql-hash-as-str "catalogs.hash"))
     certname]
    (comp first sql/result-set-seq)))
 
-;; `historical-catalogs-limit` is used for toggling historical catalog
-;; storage, this is configurable and PE only.
-(def historical-catalogs-limit (atom 0))
-
-;; `store-catalogs-jsonb-columns?` is used for toggling storage of the resources
-;; and edges jsonb blobs for catalogs. These blobs are used in PE only and this
-;; variable is meant to only be set to true in PE. This exists so that we can
-;; store the jsonb columns idependently from storing historical catalogs. This
-;; way a user can turn off historical catalogs and the PE only views still work.
-(def store-catalogs-jsonb-columns? (atom false))
-
 (defn munge-edges-for-storage [edges]
   (->> edges
        (map (fn [{:keys [source target relationship]}]
@@ -359,7 +362,7 @@
        (map (partial merge {:file nil :line nil}))
        sutils/munge-jsonb-for-storage))
 
-(pls/defn-validated catalog-row-map
+(s/defn catalog-row-map
   "Creates a row map for the catalogs table, optionally adding envrionment when it was found"
   [hash
    {:keys [edges
@@ -369,22 +372,21 @@
            transaction_uuid
            catalog_uuid
            environment
-           producer_timestamp]} :- catalog-schema
+           producer_timestamp
+           producer]} :- catalog-schema
    received-timestamp :- pls/Timestamp]
-  (let [catalogs-jsonb? @store-catalogs-jsonb-columns?]
-    {:hash (sutils/munge-hash-for-storage hash)
-     :edges (when catalogs-jsonb? (munge-edges-for-storage edges))
-     :resources (when catalogs-jsonb? (munge-resources-for-storage (vals resources)))
-     :catalog_version  version
-     :transaction_uuid (sutils/munge-uuid-for-storage transaction_uuid)
-     :catalog_uuid (sutils/munge-uuid-for-storage catalog_uuid)
-     :timestamp (to-timestamp received-timestamp)
-     :code_id code_id
-     :environment_id (ensure-environment environment)
-     :producer_timestamp (to-timestamp producer_timestamp)
-     :api_version 1}))
+  {:hash (sutils/munge-hash-for-storage hash)
+   :catalog_version  version
+   :transaction_uuid (sutils/munge-uuid-for-storage transaction_uuid)
+   :catalog_uuid (sutils/munge-uuid-for-storage catalog_uuid)
+   :timestamp (to-timestamp received-timestamp)
+   :code_id code_id
+   :environment_id (ensure-environment environment)
+   :producer_timestamp (to-timestamp producer_timestamp)
+   :producer_id (ensure-producer producer)
+   :api_version 1})
 
-(pls/defn-validated update-catalog-metadata!
+(s/defn update-catalog-metadata!
   "Given some catalog metadata, update the db"
   [id :- Long
    hash :- String
@@ -394,7 +396,7 @@
                 (catalog-row-map hash catalog received-timestamp)
                 ["id=?" id]))
 
-(pls/defn-validated add-catalog-metadata!
+(s/defn add-catalog-metadata!
   "Given some catalog metadata, persist it in the db. Returns a map of the
   inserted data including any autogenerated columns."
   [hash :- String
@@ -405,7 +407,7 @@
                        (assoc (catalog-row-map hash catalog received-timestamp)
                               :certname certname))))
 
-(pls/defn-validated resources-exist? :- #{String}
+(s/defn resources-exist? :- #{String}
   "Given a collection of resource-hashes, return the subset that
    already exist in the database."
   [resource-hashes :- #{String}]
@@ -424,7 +426,7 @@
 ;;resource-ref->resource-schema, but there are a lot of tests that
 ;;have incorrect data. When examples.clj and tests get fixed, this
 ;;should be changed to the correct schema
-(pls/defn-validated catalog-resources
+(s/defn catalog-resources
   "Returns the resource hashes keyed by resource reference"
   [certname-id :- Long]
   (jdbc/query-with-resultset
@@ -438,7 +440,7 @@
        (zipmap (map #(select-keys % [:type :title]) rss)
                (jdbc/convert-result-arrays set rss))))))
 
-(pls/defn-validated new-params-only
+(s/defn new-params-only
   "Returns a map of not persisted parameters, keyed by hash"
   [persisted-params :- #{String}
    refs-to-resources :- resource-ref->resource-schema
@@ -450,14 +452,14 @@
                    (assoc acc resource-hash parameters))))
              {} refs-to-resources))
 
-(pls/defn-validated insert-records*
+(s/defn insert-records*
   "Nil/empty safe insert-records, see java.jdbc's insert-records for more "
   [table :- s/Keyword
    record-coll :- [{s/Keyword s/Any}]]
   (when (seq record-coll)
     (apply jdbc/insert! table record-coll)))
 
-(pls/defn-validated add-params!
+(s/defn add-params!
   "Persists the new parameters found in `refs-to-resources` and populates the
    resource_params_cache."
   [refs-to-resources :- resource-ref->resource-schema
@@ -471,14 +473,17 @@
     (insert-records*
      :resource_params_cache
      (map (fn [[resource-hash params]]
-            {:resource (sutils/munge-hash-for-storage resource-hash) :parameters (when params (sutils/db-serialize params))})
+            {:resource (sutils/munge-hash-for-storage resource-hash)
+             :parameters (some-> params sutils/munge-jsonb-for-storage)})
           new-params))
 
     (insert-records*
      :resource_params
      (for [[resource-hash params] new-params
            [k v] params]
-       {:resource (sutils/munge-hash-for-storage resource-hash) :name (name k) :value (sutils/db-serialize v)}))))
+       {:resource (sutils/munge-hash-for-storage resource-hash)
+        :name (name k)
+        :value (sutils/db-serialize v)}))))
 
 (def resource-ref?
   "Returns true of the map is a resource reference"
@@ -491,7 +496,7 @@
     (update-in resource [:tags] sutils/to-jdbc-varchar-array)
     resource))
 
-(pls/defn-validated insert-catalog-resources!
+(s/defn insert-catalog-resources!
   "Returns a function that accepts a seq of ref keys to insert"
   [certname-id :- Long
    refs-to-hashes :- {resource-ref-schema String}
@@ -516,7 +521,7 @@
                 :line line})))
           refs-to-insert))))
 
-(pls/defn-validated delete-catalog-resources!
+(s/defn delete-catalog-resources!
   "Returns a function accepts old catalog resources that should be deleted."
   [certname-id :- Long]
   (fn [refs-to-delete]
@@ -559,7 +564,7 @@
                (assoc-in acc [k :resource] (get refs-to-hashes k)))
              refs-to-resources refs-to-resources))
 
-(pls/defn-validated update-catalog-resources!
+(s/defn update-catalog-resources!
   "Returns a function accepting keys that were the same from the old resources and the new resources."
   [certname-id :- Long
    refs-to-hashes :- {resource-ref-schema String}
@@ -586,7 +591,7 @@
   [resource]
   (dissoc resource :parameters))
 
-(pls/defn-validated add-resources!
+(s/defn add-resources!
   "Persist the given resource and associate it with the given catalog."
   [certname-id :- Long
    refs-to-resources :- resource-ref->resource-schema
@@ -601,7 +606,7 @@
                     (insert-catalog-resources! certname-id refs-to-hashes diffable-resources)
                     (update-catalog-resources! certname-id refs-to-hashes diffable-resources old-resources)))))
 
-(pls/defn-validated catalog-edges-map
+(s/defn catalog-edges-map
   "Return all edges for a given catalog id as a map"
   [certname :- String]
   ;; Transform the result-set into a map with [source,target,type] as the key
@@ -616,7 +621,7 @@
    #(zipmap (map vals (sql/result-set-seq %))
             (repeat nil))))
 
-(pls/defn-validated delete-edges!
+(s/defn delete-edges!
   "Delete edges for a given certname.
 
   Edges must be either nil or a collection of lists containing each element
@@ -640,7 +645,7 @@
                    (sutils/bytea-escape target)
                    type])))
 
-(pls/defn-validated insert-edges!
+(s/defn insert-edges!
   "Insert edges for a given certname.
 
   Edges must be either nil or a collection of lists containing each element
@@ -662,7 +667,7 @@
       (update! (:catalog-volatility performance-metrics) (count rows))
       (apply jdbc/insert! :edges rows))))
 
-(pls/defn-validated replace-edges!
+(s/defn replace-edges!
   "Persist the given edges in the database
 
   Each edge is looked up in the supplied resources map to find a
@@ -688,7 +693,7 @@
                    #(delete-edges! certname %)
                    identity)))
 
-(pls/defn-validated update-existing-catalog
+(s/defn update-existing-catalog
   "When a new incoming catalog has the same hash as an existing catalog, update
    performance-metrics and the transaction id for the new catalog"
   [catalog-id :- Long
@@ -699,7 +704,7 @@
   (time! (:catalog-hash-match performance-metrics)
          (update-catalog-metadata! catalog-id hash catalog received-timestamp)))
 
-(pls/defn-validated update-catalog-associations!
+(s/defn update-catalog-associations!
   "Adds/updates/deletes the edges and resources for the given certname"
   [certname-id :- Long
    {:keys [resources edges certname]} :- catalog-schema
@@ -709,7 +714,7 @@
   (time! (:add-edges performance-metrics)
          (replace-edges! certname edges refs-to-hashes)))
 
-(pls/defn-validated replace-existing-catalog
+(s/defn replace-existing-catalog
   "New catalogs for a given certname needs to have their metadata, resources and
   edges updated."
   [certname-id :- Long
@@ -725,7 +730,7 @@
          (update-catalog-metadata! catalog-id hash catalog received-timestamp)
          (update-catalog-associations! certname-id catalog refs-to-hashes)))
 
-(pls/defn-validated add-new-catalog
+(s/defn add-new-catalog
   "Creates new catalog metadata and adds the proper associations for the edges and resources"
   [certname-id :- Long
    hash :- String
@@ -735,50 +740,9 @@
   (inc! (:updated-catalog performance-metrics))
   (time! (:add-new-catalog performance-metrics)
          (let [catalog-id (:id (add-catalog-metadata! hash catalog received-timestamp))]
-           (jdbc/insert! :latest_catalogs {:certname_id certname-id :catalog_id catalog-id})
            (update-catalog-associations! certname-id catalog refs-to-hashes))))
 
-(pls/defn-validated add-catalog!
-  "Persist the supplied catalog in the database, returning its
-   similarity hash."
-  [{:keys [producer_timestamp resources certname] :as catalog} :- catalog-schema
-   received-timestamp :- pls/Timestamp
-   historical-catalogs-limit]
-  (time! (:replace-catalog performance-metrics)
-         (jdbc/with-db-transaction []
-           (let [hash (time! (:catalog-hash performance-metrics)
-                             (shash/catalog-similarity-hash catalog))
-                 {certname-id :certname_id
-                  stored-hash :catalog_hash
-                  latest-producer-timestamp :producer_timestamp} (latest-catalog-metadata certname)]
-             (inc! (:updated-catalog performance-metrics))
-             (time! (:add-new-catalog performance-metrics)
-                    (time! (get performance-metrics
-                                (if (= stored-hash hash)
-                                  (do (inc! (:duplicate-catalog performance-metrics))
-                                      :catalog-hash-match)
-                                  :catalog-hash-miss))
-                           (let [catalog-id (:id (add-catalog-metadata! hash catalog received-timestamp))]
-
-                             (when-not (some-> latest-producer-timestamp
-                                               (.after (to-timestamp producer_timestamp)))
-                               (let [refs-to-hashes (time! (:resource-hashes performance-metrics)
-                                                           (kitchensink/mapvals shash/resource-identity-hash resources))]
-                                 (if-not latest-producer-timestamp
-                                   (jdbc/insert! :latest_catalogs {:certname_id certname-id :catalog_id catalog-id})
-                                   (jdbc/update! :latest_catalogs {:catalog_id catalog-id} ["certname_id=?" certname-id]))
-                                 (update-catalog-associations! certname-id catalog refs-to-hashes)))
-                             (jdbc/delete! :catalogs
-                                           [(str "certname = ? AND "
-                                                 "id NOT IN (SELECT id FROM catalogs "
-                                                 "           WHERE certname=?"
-                                                 "           ORDER BY producer_timestamp DESC LIMIT ?)")
-                                            certname
-                                            certname
-                                            historical-catalogs-limit]))))
-             hash))))
-
-(pls/defn-validated replace-catalog!
+(s/defn replace-catalog!
   "Persist the supplied catalog in the database, returning its
    similarity hash."
   ([catalog :- catalog-schema]
@@ -809,20 +773,6 @@
                     (replace-existing-catalog certname-id catalog-id hash catalog refs-to-hashes received-timestamp))))
               hash)))))
 
-(pls/defn-validated store-catalog!
-  "Persist the supplied catalog in the database, returning its similarity hash.
-   On PE we alter the atom 'historical-catalog-limit' to be >0 and we will store
-   the catalog in database alongside the other catalogs for the node. On FOSS we
-   will not store historical catalogs and will replace the existing catalog for
-   node."
-  [catalog :- catalog-schema
-   received-timestamp :- pls/Timestamp]
-  (let [historical-catalogs-limit @historical-catalogs-limit]
-    (if (> historical-catalogs-limit 0)
-      (add-catalog! catalog received-timestamp historical-catalogs-limit)
-      (replace-catalog! catalog received-timestamp))))
-
-
 (defn catalog-hash-for-certname
   "Returns the hash for the `certname` catalog"
   [certname]
@@ -1063,7 +1013,7 @@
   "Given a certname and a map of fact names to values, store records for those
   facts associated with the certname."
   ([fact-data] (add-facts! fact-data true))
-  ([{:keys [certname values environment timestamp producer_timestamp]
+  ([{:keys [certname values environment timestamp producer_timestamp producer]
      :as fact-data} :- facts-schema
     include-hash? :- s/Bool]
    (jdbc/with-db-transaction []
@@ -1073,11 +1023,11 @@
        {:certname certname
         :timestamp (to-timestamp timestamp)
         :environment_id (ensure-environment environment)
-        :producer_timestamp (to-timestamp producer_timestamp)}
+        :producer_timestamp (to-timestamp producer_timestamp)
+        :producer_id (ensure-producer producer)}
        (when include-hash?
          {:hash (sutils/munge-hash-for-storage
-                 (shash/generic-identity-hash
-                  (dissoc fact-data :timestamp :producer_timestamp)))})))
+                 (shash/fact-identity-hash fact-data))})))
     ;; Ensure that all the required paths and values exist, and then
     ;; insert the new facts.
     (let [paths-and-valuemaps (facts/facts->paths-and-valuemaps values)
@@ -1095,7 +1045,7 @@
   "Given a certname, querys the DB for existing facts for that
    certname and will update, delete or insert the facts as necessary
    to match the facts argument. (cf. add-facts!)"
-  [{:keys [certname values environment timestamp producer_timestamp] :as fact-data}
+  [{:keys [certname values environment timestamp producer_timestamp producer] :as fact-data}
    :- facts-schema]
 
   (jdbc/with-db-transaction []
@@ -1143,8 +1093,8 @@
                    {:timestamp (to-timestamp timestamp)
                     :environment_id (ensure-environment environment)
                     :producer_timestamp (to-timestamp producer_timestamp)
-                    :hash (-> (dissoc fact-data :timestamp :producer_timestamp)
-                              shash/generic-identity-hash
+                    :hash (-> fact-data
+                              shash/fact-identity-hash
                               sutils/munge-hash-for-storage)}
                    ["id=?" factset-id]))))
 
@@ -1183,12 +1133,21 @@
       (reverse containment-path)))))
 
 (def store-resources-column? (atom false))
+
 (defn maybe-resources
   [row-map]
   (if @store-resources-column?
     row-map
     (dissoc row-map :resources)))
 
+(def store-corrective-change? (atom false))
+
+(defn maybe-corrective-change
+  [row-map]
+  (if @store-corrective-change?
+    row-map
+    (dissoc row-map :corrective_change)))
+
 (defn maybe-environment
   "This fn is most to help in testing, instead of persisting a value of
   nil, just omit it from the row map. For tests that are running older versions
@@ -1227,10 +1186,10 @@
    received-timestamp :- pls/Timestamp
    update-latest-report? :- s/Bool]
   (time! (:store-report performance-metrics)
-         (let [{:keys [puppet_version certname report_format configuration_version
+         (let [{:keys [puppet_version certname report_format configuration_version producer
                        producer_timestamp start_time end_time transaction_uuid environment
                        status noop metrics logs resources resource_events catalog_uuid
-                       code_id cached_catalog_status]
+                       code_id cached_catalog_status noop_pending corrective_change]
                 :as report} (normalize-report orig-report)
                 report-hash (shash/report-identity-hash report)]
            (jdbc/with-db-transaction []
@@ -1243,11 +1202,14 @@
                             :metrics (sutils/munge-jsonb-for-storage metrics)
                             :logs (sutils/munge-jsonb-for-storage logs)
                             :resources (sutils/munge-jsonb-for-storage resources)
+                            :corrective_change corrective_change
                             :noop noop
+                            :noop_pending noop_pending
                             :puppet_version puppet_version
                             :certname certname
                             :report_format report_format
                             :configuration_version configuration_version
+                            :producer_id (ensure-producer producer)
                             :producer_timestamp producer_timestamp
                             :start_time start_time
                             :end_time end_time
@@ -1257,14 +1219,16 @@
                    [{report-id :id}] (->> row-map
                                           maybe-environment
                                           maybe-resources
+                                          maybe-corrective-change
                                           (jdbc/insert! :reports))
-                   assoc-ids #(assoc %
-                                     :report_id report-id
-                                     :certname_id certname-id)]
+                   adjust-event-metadata #(-> %
+                                              (assoc :report_id report-id
+                                                     :certname_id certname-id)
+                                              maybe-corrective-change)]
                (when-not (empty? resource_events)
                  (->> resource_events
                       (sp/transform [sp/ALL :containment_path] #(some-> % sutils/to-jdbc-varchar-array))
-                      (map assoc-ids)
+                      (map adjust-event-metadata)
                       (apply jdbc/insert! :resource_events)))
                (when update-latest-report?
                  (update-latest-report! certname)))))))
@@ -1364,8 +1328,7 @@
            "update certnames set expired = ?"
            "  where id in"
            "    (select c.id from certnames c
-                     left outer join latest_catalogs lcats on lcats.certname_id = c.id
-                     left outer join catalogs cats on cats.id = lcats.catalog_id
+                     left outer join catalogs cats on cats.certname = c.certname
                      left outer join factsets fs on c.certname = fs.certname
                      left outer join reports r on c.latest_report_id = r.id
                    where c.deactivated is null
